#include <TFT_eSPI.h>
#include <SPI.h>
TFT_eSPI tft = TFT_eSPI();      // Display
// Dimensions écran (320 x 240)
const int TFT_MINX =   0;
const int TFT_MAXX = 280;
const int TFT_MINY =   0;
const int TFT_MAXY = 240;
#define Nombre_Couleurs 6
enum colors {RED, GREEN, BLUE, GRAY, PAL1, PAL2};
byte Palette = 1;

// Touch
boolean wastouched = true;

// Paramètres Mandel
const float Mandel_MINX = -2.1;
const float Mandel_MAXX =  0.7;
const float Mandel_MINY = -1.2;
const float Mandel_MAXY =  1.2;
float Current_MINX = Mandel_MINX;
float Current_MINY = Mandel_MINY;
float Current_MAXX = Mandel_MAXX;
float Current_MAXY = Mandel_MAXY;
float Prev_MINX = Current_MINX;
float Prev_MINY = Current_MINY;
float Prev_MAXX = Current_MAXX;
float Prev_MAXY = Current_MAXY;
float Jcx = 0.3;
float Jcy = 0.01;
int itermin = 100;
int itermax = 0;
boolean Mandel = true;
byte Puissance = 2;
#define Num_puiss 6
#define Max_iter 64
#define Escape 100
char texte[40];

void Draw_Mandel () {
  unsigned long temps = millis();
  float deltaX = (Current_MAXX - Current_MINX) / (TFT_MAXX - TFT_MINX - 1);
  float deltaY = (Current_MAXY - Current_MINY) / (TFT_MAXY - TFT_MINY - 1);
  for (int i = TFT_MINX; i < TFT_MAXX; i++) { // Premier tracé à blanc
    float cx = Current_MINX + i * deltaX;
    for (int j = TFT_MINY; j < TFT_MAXY; j++) {
      float cy = Current_MINY + j * deltaY;
      int iteration = Calcul_Mandel (cx, cy);
      if (iteration < itermin && iteration > 0) itermin = iteration;
      if (iteration > itermax) itermax = iteration;
      uint32_t color = Couleur(iteration, 0, Max_iter);
      tft.drawPixel(i, j, color);
    }
  }
  temps = millis() - temps;
  if (!Mandel) {
    tft.setTextSize(1);
    tft.setTextColor(TFT_WHITE);
    sprintf (texte, "Coefficient : ( %f , %f )", Jcx, Jcy);
    tft.drawString(texte, 25, 5);
    tft.setTextSize(2);
  }
  Serial.print ("Temps de calcul : ");
  Serial.print(temps);
  Serial.println(" ms.");
}

int Calcul_Mandel (float cx, float cy) {
  float x1 = 0;
  float y1 = 0;
  float x2 = 0;
  float y2 = 0;
  if (!Mandel) {
    x1 = cx;
    y1 = cy;
    x2 = cx;
    y2 = cy;
    cx = Jcx;
    cy = Jcy;
  }
  float temp = 0;
  int iteration = 0;

  while (true) {
    if (iteration >= Max_iter) return Max_iter;
    if (x2 * x2 + y2 * y2 > Escape) return iteration;
    for (int i = 0; i < Puissance - 1; i++) { // Calcul z**n
      temp = x1 * x2 - y1 * y2;
      y2 = x1 * y2 + x2 * y1;
      x2 = temp;
    }
    x1 = x2 + cx;
    y1 = y2 + cy;
    x2 = x1;
    y2 = y1;
    ++iteration;
  } // Fin Mandelbrot
}

uint32_t Couleur(int i, int mini, int maxi) {
  byte red = 0;
  byte green = 0;
  byte blue = 0;
  // color picker : https://www.w3schools.com/colors/colors_picker.asp
  byte col = 255 - map(i, mini, maxi, 0, 255); // de 0 à 255
  switch (Palette) {
    case RED :
      red   = col;
      green = 0;
      blue  = 0;
      break;
    case GREEN :
      red   = 0;
      green = col % 64;
      blue  = 0;
      break;
    case BLUE :
      red   = 0;
      green = 0;
      blue  = col % 32;
      break;
    case GRAY :
      red   = col % 50;
      green = red * 2;
      blue  = red;
      break;
    case PAL1 :
      red   = col % 60;
      green = col % 40;
      blue  = col % 10;
      break;
    case PAL2 : // passage bleu - vert - rouge
      red   = 0;
      green = 0;
      blue  = 0;
      if (col < 33) {
        red  = col;
      } else if (col < 200) {
        green  = col % 64; // / 4;
      } else {
        blue  = col - 200;
      }
      break;
  }
  return red << 11 | green << 5 | blue; //tft.color565(red, green, blue);
}

void Init_Ecran() {
  tft.fillRect(TFT_MAXX,  0, 40, 240, TFT_WHITE);
  tft.setTextColor(TFT_BLACK);
  tft.setTextSize(2);
  tft.drawString("R", TFT_MAXX + 15, 15);
  tft.drawString("U", TFT_MAXX + 15, 55);
  tft.drawNumber(2 + (Puissance - 1) % Num_puiss, TFT_MAXX + 15, 175);
  if (Mandel) { // Tracé Mandelbrot
    tft.drawString("J", TFT_MAXX + 15, 135);
  } else { // Tracé Julia
    tft.drawString("M", TFT_MAXX + 15, 135);
    tft.drawString("C", TFT_MAXX + 15, 215);
  }
  switch (Palette) {
    case RED :
      tft.fillRect(TFT_MAXX, 80, 40, 40, TFT_GREEN);
      break;
    case GREEN :
      tft.fillRect(TFT_MAXX, 80, 40, 40, TFT_BLUE);
      break;
    case BLUE :
      tft.fillRect(TFT_MAXX, 80, 40, 40, TFT_DARKGREY);
      break;
    case GRAY :
      tft.fillRect(TFT_MAXX, 80, 40, 40, TFT_YELLOW);
      tft.drawNumber(1, TFT_MAXX + 15, 95);
      break;
    case PAL1 :
      tft.fillRect(TFT_MAXX, 80, 40, 40, TFT_YELLOW);
      tft.drawNumber(2, TFT_MAXX + 15, 95);
      break;
    case PAL2 :
      tft.fillRect(TFT_MAXX, 80, 40, 40, TFT_RED);
      break;
  }
  for (int i = 0; i < 6; i++) {
    tft.drawRect(TFT_MAXX, i * 40, 40, 40, TFT_BLACK);
  }
}

void VisuZone (int i, int j) {
  tft.drawCircle(i, j, 4, TFT_WHITE);
  tft.drawRect(i - (TFT_MAXX - TFT_MINX) / 10, j - (TFT_MAXY - TFT_MINY) / 10,
               (TFT_MAXX - TFT_MINX) / 5, (TFT_MAXY - TFT_MINY) / 5, TFT_WHITE);
}

void CalculZone (int xtft, int ytft) {
  float coefX = (Current_MAXX - Current_MINX) / (TFT_MAXX - TFT_MINX);
  float Mandel_centerX = Current_MINX + (xtft - TFT_MINX) * coefX;
  float coefY = (Current_MAXY - Current_MINY) / (TFT_MAXY - TFT_MINY);
  float Mandel_centerY = Current_MINY + (ytft - TFT_MINY) * coefY;
  float DeltaX = (Current_MAXX - Current_MINX) / 10.;
  float DeltaY = (Current_MAXY - Current_MINY) / 10.;
  Prev_MINX = Current_MINX;
  Prev_MINY = Current_MINY;
  Prev_MAXX = Current_MAXX;
  Prev_MAXY = Current_MAXY;
  Current_MINX = Mandel_centerX - DeltaX;
  Current_MAXX = Mandel_centerX + DeltaX;
  Current_MINY = Mandel_centerY - DeltaY;
  Current_MAXY = Mandel_centerY + DeltaY;
}

void ChoixJuliaC() {
  uint16_t xtft, ytft;
  tft.fillRect(0,  0, TFT_MAXX, TFT_MAXY, TFT_WHITE);
  tft.setTextSize(1);
  tft.setTextColor(TFT_BLACK);
  tft.drawString("Choisir le coefficient C", 75, 30);
  for (int i = 0; i < 4; i++) {
    tft.drawFastVLine(i * TFT_MAXX / 4, 0, TFT_MAXY, TFT_BLACK);
    tft.drawFastHLine(0, i * TFT_MAXY / 4, TFT_MAXX, TFT_BLACK);
  }
  tft.drawNumber(-2, 5, TFT_MAXY / 2 + 5);
  tft.drawNumber(-1, TFT_MAXX / 5 + 15, TFT_MAXY / 2 + 5);
  tft.drawNumber( 1, 4 * TFT_MAXX / 5, TFT_MAXY / 2 + 5);
  tft.drawNumber( 2, 5 * TFT_MAXX / 5 - 10, TFT_MAXY / 2 + 5);
  tft.drawNumber(-1, TFT_MAXX / 2 + 5, TFT_MAXY - 15);
  tft.drawNumber( 0, TFT_MAXX / 2 + 5, TFT_MAXY / 2 + 5);
  tft.drawNumber( 1, TFT_MAXX / 2 + 5, 5);

  boolean istouched = false;
  while (!istouched) istouched = tft.getTouch(&xtft, &ytft);
  delay(200);
  istouched = false;
  while (!istouched) istouched = tft.getTouch(&xtft, &ytft);
  float coefX = 4.0 / (TFT_MAXX - TFT_MINX);
  float coefY = 2.0 / (TFT_MAXY - TFT_MINY);
  Jcx = -2.0 + xtft * coefX;
  Jcy =  1.0 - ytft * coefY;
  tft.fillCircle(xtft, ytft, 3, TFT_BLUE);
  sprintf (texte, "Coefficient : ( %f , %f )", Jcx, Jcy);
  tft.drawString(texte, 25, 50);
  delay(3000);
  float dx = 0.7;
  Current_MINX = Mandel_MINX + dx;
  Current_MINY = Mandel_MINY;
  Current_MAXX = Mandel_MAXX + dx;
  Current_MAXY = Mandel_MAXY;
}

void Gestion_bouton (int y) {
  // Boutons 40 x 40 pixels
  if (y < 40) { // Bouton RAZ
    Serial.println ("Remise a zero");
    float dx = 0.0;
    if (!Mandel) dx = 0.7;
    if (Mandel && Puissance != 2) dx = 0.65;
    Current_MINX = Mandel_MINX + dx;
    Current_MINY = Mandel_MINY;
    Current_MAXX = Mandel_MAXX + dx;
    Current_MAXY = Mandel_MAXY;
  } else if (y < 80) { // Bouton Undo
    Serial.println ("Precedent Zoom");
    Current_MINX = Prev_MINX;
    Current_MINY = Prev_MINY;
    Current_MAXX = Prev_MAXX;
    Current_MAXY = Prev_MAXY;
  } else if (y < 120) { // Bouton couleur
    Palette = (Palette + 1) % Nombre_Couleurs;
  } else if (y < 160) { // Choix Mandebrot / Julia
    Mandel = !Mandel;
    float dx = 0;
    if (!Mandel) dx = 0.7;
    Current_MINX = Mandel_MINX + dx;
    Current_MINY = Mandel_MINY;
    Current_MAXX = Mandel_MAXX + dx;
    Current_MAXY = Mandel_MAXY;
  } else if (y < 200) { // Paramètres
    Puissance = 2 + (Puissance - 1) % Num_puiss;
    float dx = 0.0;
    if (Puissance != 2) dx = 0.65;
    Current_MINX = Mandel_MINX + dx;
    Current_MINY = Mandel_MINY;
    Current_MAXX = Mandel_MAXX + dx;
    Current_MAXY = Mandel_MAXY;
  } else { // Si Julia : choix paramètre
    ChoixJuliaC();
  }
  Init_Ecran();
  Draw_Mandel();
}

void Infos() {
  tft.fillScreen(TFT_WHITE);
  tft.setTextColor(TFT_BLACK, TFT_WHITE);
  tft.setTextSize(2);
  tft.drawString("Mandelbrot - Julia", 25, 15);
  tft.setTextSize(1);
  byte y0 = 40;
  byte dy = 20;
  tft.drawString("Boutons de droite :", 25, y0 + 2 * dy);
  tft.drawString("R : Reset (full scale)", 25, y0 + 3 * dy);
  tft.drawString("U : Undo dernier zoom", 25, y0 + 4 * dy);
  tft.drawString("M/J : Toggle Mandelbrot <--> Julia", 25, y0 + 5 * dy);
  tft.drawString("Couleur : Diverses palettes", 25, y0 + 6 * dy);
  tft.drawString("Nombre : Exposant des calculs", 25, y0 + 7 * dy);
  tft.drawString("C : Choix du coefficient pour Julia", 25, y0 + 8 * dy);
  delay (6000);
}

void setup() {
  Serial.begin(115200);
  // Setup the LCD
  tft.init();
  tft.setRotation(3);
  // Init touchscreen
  //  uint16_t calData[5] = { 179, 3522, 335, 3434, 1 }; // Rotation 1
  uint16_t calData[5] = { 195, 3500, 335, 3410, 7 }; // Rotation 3
  tft.setTouch(calData);
  Infos();
  Init_Ecran();
  Draw_Mandel();
}

void loop() {
  uint16_t xtft, ytft;
  boolean istouched = tft.getTouch(&xtft, &ytft);
  if (istouched) {
    if (!wastouched) {
      if (xtft > TFT_MAXX) {  // Appui sur un des boutons
        Gestion_bouton (ytft);
      } else {
        VisuZone (xtft, ytft);
        CalculZone (xtft, ytft);
        Draw_Mandel();
      }
    }
  }
  wastouched = istouched;
}

